# -*- coding: utf-8 -*-
import datetime
import json

# Vendor Library
from yadic.container import Injectable
from django.db.transaction import commit_on_success

# Project Library
from kinder.core.audit_log.helpers import get_model
from kinder.webservice.spyne_ws.exceptions import SpyneException
from kinder.core.declaration.models import Declaration, DeclarationPrivilege
from kinder.core.declaration_status.models import (
    DeclarationStatus, DSS)
from kinder.core.children.models import ChildrenDelegate, Delegate
from kinder.core.helpers import recursive_getattr
from kinder.provider import features

from concentrator.models import ChangeStatus, ChangeSource

from lipetsk_specifics.change import (
    DeclarationPrivilegeChangeHelper, ChangeHelper, map_changes, StorageHelper)
from lipetsk_specifics.rules import (WorkTypeRules, DelegateWhoHaveBenefits)
from lipetsk_specifics.models import(
    PrivilegeConfirmationAttributes, PrivilegeOwnerEnum)

from .spyne_objects_proxy import (
    ApplicationRulesDataProxy, DocumentReferencesDataProxy)
from .django_objects_proxy import (
    LipetskDelegateProxy,
    LipetskChildrenProxy,
    DelegatePrivilegeProxy,
    PrivConfirmAttrsProxy,
    PrivilegeProxy,
    DocExtraInfoProxy
)
from .helpers import (
    get_decl_priv, get_priv_data, get_delegate,
    _get_document_reference_from_ctx, DocumentWorker,
    get_mo_from_request, get_declaration_units_from_request,
    get_declaration_units, get_desire_date, get_health_need)
from .entities import UpdateApplicationResponse, GetApplicationResponse


class LipetskNewApplicationProxy(object):
    """
    Прокси объект для метода NewApplication
    """

    def __init__(self, request, binary_data, req_code):

        self._external_id = request.ExternalId
        self._submit_date = request.SubmitDate
        self._entry_date = request.EntryDate
        self._adaptation_program_type = request.AdaptationProgramType
        self._delegate_data = request.Applicant
        self._child_data = request.DeclaredPerson
        self._edu_organizations = request.EduOrganizations
        self._binary_data = binary_data
        self._request_code = req_code
        self._document_references = _get_document_reference_from_ctx(
            request)

    def _create_delegate_for_priv(self, delegate_type, who_type, children,
                                  delegate):
        """
        Проверяем по ФИ и снилсу не привязан ли уже этот представитель
        к ребенку, если нет, то создаем нового, иначе возвращаем существующего
        @param delegate: уже созданные представитель в заявке
        @param who_type: Тип описывающей владельца льготы
        @param delegate_type: Тип описывающей представителя ребенка
        """
        new_delegate = None
        map_compare = ['LastName', 'FirstName', 'Snils']
        if all(getattr(delegate_type, field_name) == getattr(who_type,
                                                             field_name)
               for field_name in map_compare):
            new_delegate = delegate
        if not new_delegate:
            new_delegate, new_delegate_extension = DelegatePrivilegeProxy(
                who_type).get()
            new_delegate.save()
            new_delegate_extension.delegate = new_delegate
            new_delegate_extension.save()
            ChildrenDelegate.objects.create(
                children=children, delegate=new_delegate)
        return new_delegate

    def _save_privileges(self, privileges, declaration, doc_number, delegate):
        """
        Согласовано с аналитикаами,
        что при подаче с портала не может быть более 1 льготы
        @param privileges:
        @param declaration:
        @param doc_number:
        @param delegate:
        @return:
        """
        if len(privileges) > 1:
            raise SpyneException(u"Не должно быть более 1 льготы с портала!!!")
        if privileges:
            owner_type = self._child_data.Benefits.WhoHaveBenefit.Type
            owner_delegate = None
            if owner_type != PrivilegeOwnerEnum.CHILDREN:
                owner_delegate = self._create_delegate_for_priv(
                    self._delegate_data,
                    self._child_data.Benefits.WhoHaveBenefit,
                    declaration.children,
                    delegate)
            for privilege in privileges:
                decl_pr = DeclarationPrivilege.objects.create(
                    declaration=declaration,
                    datetime=datetime.datetime.now(),
                    privilege=privilege,
                    doc_number=doc_number)
                data = dict(
                    privilege_owner=owner_type,
                    declaration_privilege=decl_pr,
                    portal=True,
                )
                if owner_delegate:
                    data['delegate'] = owner_delegate
                PrivilegeConfirmationAttributes.objects.create(**data)

    def _create_delegate(self):
        return LipetskDelegateProxy(self._delegate_data).get()

    def _create_children(self):
        return LipetskChildrenProxy(
            self._child_data, self._delegate_data,
            self._adaptation_program_type).get()

    def _create_benefits(self):
        result = []
        doc_number = (self._child_data.Benefits.BenefitsDocInfo
                      if self._child_data.Benefits else None)
        if self._child_data.Benefits and self._child_data.Benefits.Benefit:
            result = PrivilegeProxy([
                (benefit.name,
                 benefit.Type)
                for benefit in self._child_data.Benefits.Benefit]).get()
        return result, doc_number

    def _get_work_type(self):
        return WorkTypeRules.get_system(self._child_data.ArrivalTimeType)

    def _create_declaration(self):
        return Declaration(
            date=self._submit_date,
            desired_date=self._entry_date,
            offer_other=self._edu_organizations.AllowOfferOther,
            client_id=self._external_id,
            comment=self._child_data.IssuerState,
            portal=1,
            work_type_id=self._get_work_type(),
            source=features.get_feature(
                'lipetsk_specifics.webservice.service.'
                'NewApplicationRequest.get_source')())

    def _attache_documents(self, declaration):
        DocumentWorker(
            declaration, self._binary_data, self._request_code).attach()

    def _save_doc_extra_info(self, declaration):
        DocExtraInfoProxy(declaration, self._document_references).save()

    @commit_on_success
    def process(self):
        if Declaration.objects.filter(client_id=self._external_id).exists():
            raise SpyneException(
                u'Заявление с ExternalId = %s уже существует' %
                self._external_id)
        declaration = self._create_declaration()
        delegate, delegate_extension = self._create_delegate()
        children = self._create_children()
        privileges, doc_number = self._create_benefits()
        mo = get_mo_from_request(self._edu_organizations, declaration)

        delegate.save()
        delegate_extension.delegate = delegate
        delegate_extension.save()
        children.save()
        ChildrenDelegate.objects.create(
            children=children, delegate=delegate)

        declaration.children = children
        declaration.mo = mo
        declaration.status = DeclarationStatus.objects.get(
            code=DSS.DUL_CONFIRMATING)
        declaration.save()

        declaration_units = get_declaration_units_from_request(
            self._edu_organizations, declaration)

        for declaration_unit in declaration_units:
            # bulk create не использую т.к. он не вызывает save
            # метода и не срабатывают сигналы post_save и pre_save
            declaration_unit.declaration = declaration
            declaration_unit.save()

        self._save_privileges(privileges, declaration, doc_number, delegate)
        self._save_doc_extra_info(declaration)
        self._attache_documents(declaration)

        return declaration.id


class GetApplicationProxy(object):
    """
    Прокси объект для метода GetApplication
        Получение данных Заявления для изменения
    """
    def __init__(self, declaration, decl_proxy, app_proxy, dou_proxy):
        assert declaration
        assert isinstance(declaration, Declaration)
        self.declaration = declaration
        self.decl_proxy = decl_proxy
        self.app_proxy = app_proxy
        self.dou_proxy = dou_proxy

    def process(self):
        conf = {
            'SubmitDate': self.declaration.date,
            'EntryDate': get_desire_date(self.declaration),
            'EducationProgramType': '',
            'AdaptationProgramType': get_health_need(self.declaration),
            'Applicant': self.app_proxy.create_spyne_obj(),
            'DeclaredPerson': self.decl_proxy.create_spyne_obj(),
            'EduOrganizationsData': self.dou_proxy.create_spyne_obj(),
            'ApplicationRules': ApplicationRulesDataProxy.create_rules_data(
                self.declaration),
            'DocumentReferences': DocumentReferencesDataProxy.get_data(
                self.declaration)
        }

        return GetApplicationResponse(**conf)


class LipetskDeclarationPrivilegeChangeHelper(
    DeclarationPrivilegeChangeHelper
):
    """
    Изменени порядок олучения изменений по льготам
    Фиксируем изменения только для льгот пришедших с портала
    """
    def _get_current_state(self, difference, changes):
        """
        возвращает список id льгот заявки
        @param difference:
        @param changes:
        @return:
        """
        if changes.state == ChangeStatus.WAIT:
            declarationprivilege_set = get_decl_priv(changes.declaration, True)
            curent_state = map(lambda x: x.privilege.id,
                               declarationprivilege_set)
        else:
            curent_state = map(lambda x: x[self.SYS_NAME],
                               filter(lambda x: self.SYS_NAME in x.keys(),
                                      difference))
            curent_state = curent_state[0]
        return curent_state

    def _create_new(self, new_list, changes):
        """
        создаем привязку льготы к заявлению
        @param new_list:
        @return:
        """
        if len(new_list) > 1:
            raise SpyneException(u"Не должно быть более 1 льготы с портала!!!")
        for priv_id in new_list:
            decl_pr, _create = DeclarationPrivilege.objects.get_or_create(
                declaration=changes.declaration,
                privilege_id=priv_id,
            )
            # Если данные по обладателю льготы уже есть,
            # то мы только изменим в нем ссылку на льготу,
            # иначе создадим заново.
            # Тк в липецке льгота одна, считаем что именно ее обладателя мы и
            # должны изменить.
            priv_atr_qs = PrivilegeConfirmationAttributes.objects.filter(
                declaration_privilege__declaration_id=changes.declaration)
            if priv_atr_qs:
                for attr in priv_atr_qs:
                    attr.declaration_privilege = decl_pr
                    attr.confirmed = False
                    attr.save()
            else:
                data = dict(
                    declaration_privilege=decl_pr,
                    portal=True,
                )
                owner_delegate = self._create_delegate_from_change(
                    changes) or get_delegate(changes.declaration)

                if owner_delegate:
                    data['delegate'] = owner_delegate
                    data['privilege_owner'] = (
                        DelegateWhoHaveBenefits.get_concetr(owner_delegate.type))
                PrivilegeConfirmationAttributes.objects.create(**data)

    def _create_delegate_from_change(self, changes):
        try:
            changes_dict = json.loads(changes.data)
        except ValueError:
            return None

        try:
            confirmation_attributes = changes_dict[
                "PrivilegeConfirmationAttributes"]
        except KeyError:
            return None

        data = {}
        for attribute in confirmation_attributes:
            for field in attribute:
                if field.startswith("delegate."):
                    _, field_name = field.split(".")
                    field_changes_info = attribute[field]
                    _, value, _ = field_changes_info
                    data[field_name] = value

        try:
            delegate = Delegate.objects.create(**data)
        except ValueError:
            return None

        return delegate


class PrivConfAttrsChangeHelper(ChangeHelper):
    u"""
    Обработчик изменений доп.аттрибутов льгот PrivilegeConfirmationAttributes.
    При проверки изменений атрибутов обладателя льготы, вжно что пришло в
    теге Type, если пришел PrivilegeOwnerEnum.DELEGATE,
    то теги данных будут заполнены и их нужно проверять на изменение,
    иаче проверяем только на изменение тип представителя

    """
    def check_diff(self, conc_object, save_object):
        if conc_object.privilege_owner != PrivilegeOwnerEnum.DELEGATE:
            instanse_model = save_object or conc_object
            field, type_value, _ = ('privilege_owner', int, 'Type')
            conc_value = recursive_getattr(conc_object,
                                           field.replace('.', '__'))
            save_value = recursive_getattr(save_object,
                                           field.replace('.', '__'))
            compare_func = map_changes.get_function_compare(type_value)
            if not compare_func(conc_value, save_value):
                self._handling_change(field, type_value, conc_value,
                                      save_value, instanse_model)
        else:
            super(PrivConfAttrsChangeHelper, self).check_diff(conc_object, save_object)

    def apply_change(self, difference, changes):
        """Если пришли данные на изменения обладателя льготы, а с такими данными
        представитель у ребенка есть(ФИО, типу ДУЛ, серии и номеру ДУЛ),
        то текущего мы удаляем, а льготу перепривязываем к уже существующему

        - применяет изменения в системе
        - возаращает список изменений в окно
        @return:
        """
        list_params = ['delegate__surname', 'delegate__firstname',
                       'delegate__patronymic',
                       'delegate__dul_type', 'delegate__dul_series',
                       'delegate__dul_number', 'delegate__id']

        updated_fields = super(PrivConfAttrsChangeHelper, self).apply_change(
            difference, changes)

        delegate_qs = changes.declaration.children.childrendelegate_set
        priv_conf_attrs = get_priv_data(changes.declaration)
        if priv_conf_attrs:
            priv_delegate_params_qs = delegate_qs.filter(
                delegate__id=priv_conf_attrs.delegate_id
            )
            if priv_delegate_params_qs.exists():
                priv_delegate_params = (
                    priv_delegate_params_qs.values(*list_params)[0]
                )
                another_delegates = delegate_qs.exclude(
                    delegate__id=priv_conf_attrs.delegate_id
                ).values(*list_params)

                for delegate_params in another_delegates:
                    if all(delegate_params[key] == priv_delegate_params[key]
                           for key in filter(
                            lambda x: x !='delegate__id', list_params)):
                        priv_conf_attrs.delegate_id = delegate_params['delegate__id']
                        priv_conf_attrs.save()
                        Delegate.objects.get(
                            id=priv_delegate_params['delegate__id']).delete()
                        break
        return updated_fields


class LipetskUpdateProxy(LipetskNewApplicationProxy):
    """
    Сравниваем только с льготами с портала
    """

    __metaclass__ = Injectable

    def __init__(self, request, declaration, binary_data, request_code,
                 **kwargs):
        self._external_id = request.ExternalId
        self._submit_date = request.SubmitDate
        self._entry_date = request.EntryDate
        self._adaptation_program_type = request.AdaptationProgramType
        self._delegate_data = request.Applicant
        self._child_data = request.DeclaredPerson
        self._edu_organizations = request.EduOrganizationsData
        self.declaration = declaration
        self.result = []
        self._binary_data = binary_data
        self._request_code = request_code
        self._document_references = _get_document_reference_from_ctx(
            request)
        self.case_number = kwargs.get("case_number", None)

    def chek_model(self, name_model, conc_data, have_data):
        """
        @param name_model: str имя модели из MAP_CHANGES
        @param conc_data: spyne object инстанс модеи воссозданный
         из данных на изменение
        @param have_data: django model инстанс моделииз хранилища
        """
        class_model = get_model(name_model)
        class_helper = StorageHelper.get_change_helper(name_model)
        ch_help = class_helper(class_model, name_model)
        ch_help.check_diff(conc_data, have_data)
        self.result.append(ch_help)

    def _get_declaration_units_list(self):
        """Возвращает список (unit_id, ord) по данным из запроса,
        прошедших всю валидацию"""
        decl_units = get_declaration_units_from_request(
            self._edu_organizations, self.declaration)
        return map(lambda x: (x.unit_id, x.ord), decl_units)

    def _create_benefits(self):
        result, doc_number = super(LipetskUpdateProxy, self)._create_benefits()
        return map(lambda x: x.id, result)

    def _save_documents(self):
        """
        Сохранение документов
        """
        DocumentWorker(
            self.declaration, self._binary_data, self._request_code).attach()
        DocExtraInfoProxy(self.declaration, self._document_references).save()

    def _create_delegate(self):
        return LipetskDelegateProxy(self._delegate_data).get()

    def _create_children(self):
        return LipetskChildrenProxy(
            self._child_data, self._delegate_data,
            self._adaptation_program_type).get()

    def _check_benefits(self):
        conc_benefits = self._create_benefits()
        declarationprivilege_set = get_decl_priv(self.declaration, True)
        benefits = map(lambda x: x.privilege.id, declarationprivilege_set)

        # 1. Основная модель привелегий
        self.chek_model('DeclarationPrivilege', conc_benefits, benefits)

        # 2. Модель доп.аттрибут о льготах (липецкая, внутри плагина)
        if self._child_data.Benefits:
            new_priv_conf_attrs = self._create_privilegeconfattributes()
            priv_conf_attrs = get_priv_data(self.declaration)
            if new_priv_conf_attrs:
                self.chek_model(
                    'PrivilegeConfirmationAttributes',
                    new_priv_conf_attrs, priv_conf_attrs)

    def _create_privilegeconfattributes(self):
        u"""Получение прокси класса по доп.аттрибутам льгот.
        """
        #В липецке на портале нельзя указать больше 1 льготы
        declaration_privilege_set = get_decl_priv(self.declaration)
        if declaration_privilege_set:
            return PrivConfirmAttrsProxy(
                self._child_data,
                self.declaration,
                declaration_privilege_set[0]
            ).get()
        else:
            privilege_id = self._child_data.Benefits.Benefit[0].name
            return PrivConfirmAttrsProxy(
                self._child_data,
                self.declaration,
                DeclarationPrivilege(
                    declaration=self.declaration,
                    privilege_id=privilege_id)
            ).get()

    def process(self):
        msg = u'OK'
        #проверем поля заявки
        self.chek_model('Declaration', self._create_declaration(),
                        self.declaration)
        #проверяем поля ребенка
        self.chek_model('Children', self._create_children(),
                        self.declaration.children)
        #проверяем поля представителя
        conc_delegate, _ = self._create_delegate()
        #тут также надеемся что родитль один и его же мы и редактируем!
        delegate = get_delegate(self.declaration)
        self.chek_model('Delegate', conc_delegate, delegate)

        # проверяем поле ЖУ
        conc_decl_units = self._get_declaration_units_list()
        sys_decl_units = get_declaration_units(self.declaration)
        self.chek_model('DeclarationUnit', conc_decl_units, sys_decl_units)

        self._check_benefits()

        result = StorageHelper.create_change(
            self.declaration, self.result,
            source=ChangeSource.UPDATE_APPLICATION,
            case_number=self.case_number)
        if not result:
            msg = u'Diff not found'

        # Сохранение документов без подтверждения
        self._save_documents()

        return UpdateApplicationResponse(**{
            'Status': msg
        })

